// Implements a dictionary's functionality

#include <stdbool.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <cs50.h>
#include <string.h>
#include <strings.h>
#include "dictionary.h"

// Represents a node in a hash table
typedef struct node
{
    char word[LENGTH + 1];
    struct node *next;
}
node;

// Number of buckets in hash table
// After trying out higher and lower prime numbers, it seems those in the 5000's range give best results (overall time achieved 0.05!)
const unsigned int N = 5153; ;

// Hash table
node *table[N];


// delcare prototype of move_cursor functions - update: I'm not using them, but keeping them in here for me as a reminder of lessons to learn ;)
bool move_cursor(node *cursor, const char *word1); // check
void move_cursortemp(node *cursor);

// Size Counter:
int word_count = 0;




// Returns true if word is in dictionary, else false
bool check(const char *word)
{

    int index = hash(word);

    // set variable cursor to the same as the one starting the linked list
    node *cursor = table[index];

    while (cursor != NULL)
    {
        // compare words case insensitively
        if (strcasecmp(cursor -> word, word) == 0)
        {
            // #found
            return true;
        }
        else if (cursor == NULL)
        {
            // means we reached the end of the linked list, so word isn't here
            return false;
        }
        else
        {
            // if not found but cursor not yet null, move ooonnn
            cursor = cursor -> next;
        }
    }
    return false;
}


// Hashes word to a number
unsigned int hash(const char *word)
{

    /*
    Reference:
    HASH FUNCTION BASED ON: https://www.hackerearth.com/practice/data-structures/hash-tables/basics-of-hash-tables/tutorial/
    I adapted the function that takes each letter in the key and multiplies it by its own order in the word,
    and then applies the modulo with a high prime number (I tried out a few and chose 5153)
    */

    // declare index and value
    int index = 0;
    int value = 0;

    // Get word length and use it to go over each letter
    int word_length = strlen(word) - 1;

    for (int i = 0; i < word_length; i++)
    {
        int lower = tolower(word[i]);

        // use each letters ASCII value and multiply by its position in the word
        value += lower * (i + 1);
    }
    // Calculate index with modulo of N
    index = value % N;

    return index;
}



// Loads dictionary into memory, returning true if successful, else false
bool load(const char *dictionary)
{
    // first of all, set all pointers in hash table to null
    for (int j = 0; j < N; j++)
    {
        table[j] = NULL;
    }

    // open file for reading
    FILE *dict = fopen(dictionary, "r");
    // Check if NULL
    if (dict == NULL)
    {
        printf("Could not open.\n");
        return false;
    }

    // define a buffer variable for each word
    char wordtmp[LENGTH + 1];

    // do until end of dictionary
    while (fscanf(dict, "%s", wordtmp) != EOF)
    {
        // Update word count with each word
        word_count++;

        // Allocate memory for a new node for each word
        node *n = malloc(sizeof(node));
        // Make sure node is not NULL
        if (n == NULL)
        {
            unload();
            return false;
        }

        // Fill new node with word and set pointer to null
        strcpy(n->word, wordtmp);
        n->next = NULL;

        // use hash function to determine where to put the node
        int index = hash(wordtmp);
        // printf("Hashvalue for word %s is %int.\n", wordtmp, index);


        // Insert into linked list
        if (table[index] == NULL)
        {
            // Make this current word node that index
            table[index] = n;
        }
        else
        {
            // first make our current "next" pointer point to the same node that the table[index] is pointing to,
            //so we preseve the linked list!
            n->next = table[index];

            // Then, we can make the node n the first:
            table[index] = n;
        }
    }

    fclose(dict);

    return true;
}



// Returns number of words in dictionary if loaded, else 0 if not yet loaded
unsigned int size(void)
{
    // word count kept tack of in global variable (increased with each word in load function)
    return word_count;
}



// Unloads dictionary from memory, returning true if successful, else false
bool unload(void)
{
    node *cursor = NULL;
    node *temp = NULL;

    // for every bucket in hash table
    for (int i = 0 ; i < N; i++)
    {

        // set cursor variable to the same that the first pointer is pointing at
        cursor = table[i];

        // Do so long as the cursor itself is not == null
        while (cursor != NULL)
        {
            // make a copy of cursor in temp
            temp = cursor;
            // move cursor to the next item in linked list
            cursor = cursor -> next;
            // free temp
            free(temp);
        }

        // recursive function
        // move_cursortemp(cursor);
    }
    return true;
}








// Failed attempt at recursion - function unused
// recursive function check
bool move_cursor(node *cursor, const char *word1)
{
    // Breakpoint
    if (strcasecmp(cursor -> word, word1) == 0)
    {
        return true;
    }
    // Else if the cursor is not pointing to null, then try next one
    else if (cursor -> next != NULL)
    {
        cursor = cursor -> next;
        // Recursively use this function
        move_cursor(cursor, word1);
        return true;
    }
    else
    {
        return false;
    }
}

// Failed attempt at resursive functions! Had an issue with double-free
// recursive function - unload
void move_cursortemp(node *cursor)
{
    // if first element is already pointing at null
    if (cursor -> next == NULL)
    {
        // simply free cursor and return
        free(cursor);
        return;
    }
    // if cursor is not pointing at null
    else if (cursor -> next != NULL)
    {
        // set temp pointer to cursor pointer and then move cursor and free temp
        node *temp = cursor;
        cursor = cursor -> next;
        free(temp);

        // call recursive function again
        move_cursortemp(cursor);
    }
}